/*
   $Id: BankBuffer.cpp 649 2011-10-31 13:57:42Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#include <cerrno>
#include "BankBuffer.h"
#include "utils.h"

/**
 * \brief The constructor
 * \param[in,out] b The black list
 * \param[in] m If true, we work in "memory_mode", ie data are indexed but not saved to a file
 * */
BankBuffer::BankBuffer(BlackList& b, bool m):
	__parameters(Parameters::Instance()),
	__bank(Bank::Instance()),
	__mask(Mask::Instance()),
	__window(__mask),
	__index_tree(IndexTree::Instance()),
	__black_list(b),
	__memory_mode(m),
	__fa_bfr(__InitFaBfr()),
	__bi_bfr(__InitBiBfr()),
	__fa_file(NULL)
{
	__ResetPos();
}

/** 
 * \brief Init SeqBfr and bi_bfr - The initialization depends on __memory_mode
*/
inline Buffer* BankBuffer::__InitFaBfr()
{
    size_t max_size = __memory_mode ? 0 : SBUFFER_MAX_SIZE;
    return new Buffer(SBUFFER_INITIAL_SIZE,max_size);
}

/** 
 * \brief Init SeqBfr and bi_bfr - The initialization depends on __memory_mode
*/
Buffer* BankBuffer::__InitBiBfr()
{
    if (__memory_mode)
        return new Buffer(SBUFFER_INITIAL_SIZE);
    else
        return NULL;
}

BankBuffer::~BankBuffer() {
    __Syncr();       // Synchronize the buffer before removing it
    
    // Give the ownership of bi_bfr and fa_bfr to bank
    // This was WITHDRAWN because we had perf pbs in g++ - See also ReadSeqMemory
    //bank.SetBfr(bi_bfr,fa_bfr);
}

/**
 * \brief Index the sequence
 * \return The number of positions recorded in the tree
 * \public
 */

void BankBuffer::IndexSequence() {
    if ( __memory_mode )
		__IndexSequenceMemory();
	else
		__IndexSequenceFile();
}

/**
 * \brief Index the sequence when memory mode is false. Called by IndexSequence
 *        - Call __Readseq for every pass (switch -p)
 *        - the fa_file file descriptor is opened and written at the first pass
 * \return The number of positions recorded in the tree
 * \private
 */
void BankBuffer::__IndexSequenceFile()
{
    string filename                  = __bank.GetFileName().GetName();
    CompressionAlgorithm compression = __bank.GetFileName().GetCompression();
    
    FILE* in;         // The input file, in fasta format
    FILE *bi_file;    // The binary index file

    char c;

    in      = FOpen (filename, "",           "r", compression);
    bi_file = FOpen (filename, BI_EXTENSION, "w", compression);

    int pass_nb=1;
    do
    {
		fprintf(stderr,"Reading the sequence : pass %d/%d\n",pass_nb,__parameters.GetIndexNbPass());
        if (pass_nb==1)
        {
			__fa_file = FOpen(filename, FA_EXTENSION, "w", compression);
            __ReadSeq(in);  // Index for this pass, reading the fasta file and dumping to fa
			FClose(in);     // For the other passes, read from the fa file

            c='\n';
            fwrite(&c,sizeof(char), 1, __fa_file); // Better terminating the file by \n
            FClose(__fa_file);           

			in = FOpen(filename, FA_EXTENSION, "r", compression); // Open the .fa file for the other passes
        }
        else
        {
            rewind(in);
            __ResetPos();
            __ReadSeq(in);  // Index for this pass, reading the fa file (now opened through in)
        };
#ifdef DEBUGLOG
        cerr << "Memory used for this pass = " << MemoryPool::Instance().MemoryUsed()/1048576 << "Mb ";
        cerr << "(allocated " << MemoryPool::Instance().MemoryAllocated()/1048576 << "Mb)\n";
        cerr << "                Nodes     = " << Node::MemoryUsed()/1048576 << "Mb\n";
        cerr << "                Leaves    = " << Leaf::MemoryUsed()/1048576 << "Mb\n";
        cerr << "                Positions = " << Pos::MemoryUsed()/1048576 << "Mb\n";
#endif

        __index_tree.WriteCompressed(bi_file);

        //cerr << pass_nb << ' ' << bank.GetNbWords() << ' ' << bank.GetLength() << ' ' << bank.GetNbPos() << '\n';
	} while ((pass_nb=__index_tree.NextPass())!=0);
	FClose(bi_file);
}

/**
 * \brief Index the sequence in memory mode. Called by IndexSequence
 * \note  Only ONE pass is possible (the -p switch cannot be used when in this mode)
 * \return The number of positions recorded in the tree
 * \private
 */

void BankBuffer::__IndexSequenceMemory()
{
    string filename                  = __bank.GetFileName().GetName();
    CompressionAlgorithm compression = __bank.GetFileName().GetCompression();
    FILE* in = FOpen(filename,"","r",compression);

    fprintf(stderr,"Reading the sequence \n");
    __ReadSeq(in);
    fclose(in);

    /*cerr << "Memory used = " << MemoryPool::Instance()->MemoryUsed()/1048576 << "Mb ";
    cerr << "(allocated " << MemoryPool::Instance()->MemoryAllocated()/1048576 << "Mb)\n";
    cerr << "       Nodes     = " << Node::MemoryUsed()/1048576 << "Mb\n";
    cerr << "       Leaves    = " << Leaf::MemoryUsed()/1048576 << "Mb\n";
    cerr << "       Positions = " << Pos::MemoryUsed()/1048576 << "Mb\n";*/

    __index_tree.WriteCompressed(__bi_bfr.get());

    // Give the ownership of bi_bfr and fa_bfr to __bank
    // WARNING - If you execute IndexSequence 2 times, plantage !
    // This should be solved with putting the next line in the destructor, but then I have very poor performances
    __bank.SetBfr(__bi_bfr,__fa_bfr);
    //locked = true;                    // Preventing from any other indexation
 
}

/** 
 * \brief Called by the __IndexSequence*** functions. The sequence file is reformatted and written to out:
 *        -# The headers and the end-of-line are removed
 *        -# Sequences in a multifasta file are separated by $$$$$$$$$$
 *        -# The first and last characters of the file is $
 *        -# The low complexity regions are written in low case
 *        -# Dust is called if necessary
 * 
 * \param in The open input file
 * \return The sequence length
 */  

void BankBuffer::__ReadSeq(FILE *in)
{
	int  c;
	int i;

	size_t Sequence_Length = 0;  /* Size of the sequences that will be indexed */
    {
		uint32_t start = 0;
		uint32_t end   = 0;
		string sequence_name;
        while ((c=fgetc(in))!=EOF)
        {
            if (c=='>')                                         // Next sequence
            {
				// Finish with the current sequence, if any
				if (sequence_name != "")
				{
					end = __GetPosition()-1;   //The end was the previous position
					__bank.PushSeq(sequence_name,start,end);
				}
				// Next sequence starts here
                sequence_name = GetSeqName(in);
                if (sequence_name=="")
                    break;                                      // Job finished

                for (i=0;i< SEQUENCE_SEPARATOR_SIZE ;i++) {                            // write SEQUENCE_SEPARATOR_SIZE '$'
                    __PutSeparator();

                if (__GetPosition() != 0 && __GetPosition()%5000000==0)
                        cerr<<"bank position "<<__GetPosition()<<endl;
                };
                if (__index_tree.GetPass()==1)
                {						
	                start = __GetPosition();              
	            } 
	            else
	            {	//Should never happen !!!!!!!
					cerr<<"Something wrong here "<<endl;	
				}
            }
            else
            {
				__Putc(c);
				if ( __index_tree.GetPass()==1 && c!='\n')
					Sequence_Length++;
                if (__GetPosition() != 0 && __GetPosition()%5000000==0)
                    cerr<<"bank position "<<__GetPosition()<<endl;
            };
        };

		if (__index_tree.GetPass()==1)
		{
			if (sequence_name != "")  // if we did not exit with the break !
			{	
				end = __GetPosition()-1;     //The end was the previous position
				__bank.PushSeq(sequence_name,start,end);
			}			  	
		  	for (i=0;i< SEQUENCE_SEPARATOR_SIZE ;i++) {                            // write SEQUENCE_SEPARATOR_SIZE '$'
         		__PutSeparator();
			}
			__bank.SetSequenceLength( Sequence_Length );						
			__bank.SetLength( __GetPosition() );	
		}     
    };

    __Syncr();	
    	
	if ( __index_tree.GetPass() == __index_tree.GetNbPass() )
		cerr<<"Sequence length = "<<__bank.GetSequenceLength()<<endl;
		
}

/** 
 * \brief If the word has already too many occurrences, black list it and delete the corresponding leaf.
 *        Else, call Leaf::NewPos
 * \param l Pointer to the leaf in which the new position will be inserted
 * \param pos The position to add to this leaf
 */
 
void BankBuffer::__NewPosition(Leaf* l,uint32_t pos)
{
	if (l->GetNbPos()> __parameters.GetBlackListMaxCount()) {
        __black_list.Insert(__window.GetWord());
		//cerr << "Frequent word, deleted " << window.GetWord().Convert() << " !! " << __bank.Position << '\n';
		//cerr << "Frequent word, deleted " << window.GetWord().Convert() << '\n';
		/*  On pourrait ici remettre a 0 la fenetre !! (attention a dinucleotide counter)  */
        __index_tree.Delete(__window);
		/*   __bank->Words-=f->nb_pos;  */
		//__bank.Words--;
		__bank.DecrNbWords();
	} else {
		/*  __bank->Words++;  */
        l->NewPos(pos,__mask.GetLength());
	}
}

/** 
 * \brief Called by __Index***(). 
 *        If the word is not already black listed, retrieve the corresponding leaf, inserting a new leaf if necessary.
 *        However, nothing is inserted if we are not in the good phylum (see the IndexTree class)
 *        If a new leaf was inserted, call __NewPosition
 * \param pos The position to add to the leaf
 */

void BankBuffer::__ReadWord(uint32_t pos)
{
    if (__black_list.LookUp(__window.GetWord())==false) {
        Leaf *l = __index_tree.Insert(__window);    /* insert the new word or update count if the word already exists  */
        if (l !=NULL)                      // may be null if wrong pass treated
            __NewPosition(l,pos);
    };
}

/** 
 * \brief Called by _Syncr when COMPLEXITY_NONE or COMPLEXITY_DUST is selected
 *        Shift a sliding window and call __ReadWord when a word is found
 */
void BankBuffer::__Index()
{
    if (__IsEmpty())
        return;

	//Warning the starting pos might be negative in order to account for filling the window while working with the first fa_bfr
	uint32_t pos = __GetStartingPos();	
	
	__window.Reset();
	
    size_t i = 0;
    while(i < __fa_bfr->size())
    {
        char c = *((char*)__fa_bfr->GetBase() + i++);
        if (Word::IsCode(c)) { 
			__window.Shift(c);        // Shift the window, computing the complexity (dinucleotides)
            if (__window.IsFull()) {
                __ReadWord(pos+i);
            };
        }           
        // Not a nucleotide: Reset the window
        else
        {
            __window.Reset();
        };
    };
}

/**
 * \brief Called by BankBuffer::__Putc(char) or by the destructor of BankBuffer, _Syncr works differently in memory_mode and not memory_mode:
 *        - If NOT in memory_mode:
 *               - Call __ComplexityMasking() : masking the sequence according to a complexity measure (entropy or number of diff dinucleotides)
 * 				 - Call __Index()             : indexing the different words of the bank
 *               - Call _Dump() if fa_file not NULL (i.e first (or only) indexing pass
 *               - Clear fa_bfr
 *        - If in memory_mode:
 *               - Call __ComplexityMasking()
 *               - Call __Index()
 */
void BankBuffer::__Syncr()
{
    
    if (__fa_bfr.get()!=NULL && !__IsEmpty())
    {
        if (!__memory_mode)       // __memory_mode false: writing the index to files
        {
            if ( __fa_file != NULL ) //(when __fa_file==NULL, the pass number is higher than 1 and the complexity management has already been done */
				__ComplexityMasking();
				
	        __Index();                   // Start the indexation 
			
            if (__fa_file != NULL)             // Dump to the fa file
                __Dump();  
        }
        else                    // __memory_mode true: writing the index to memory
        {	
			__ComplexityMasking();       // Masking if necessary
            __Index();                   // Start the indexation
        };
        
        __Clear();                        // Clear the buffer in any case
    };
}


/** 
 * \brief Called by __Syncr() according to the complexity flag (none, or dust)
 *        call the corresponding complexity evaluation procedure and associated masking routine
 */
void BankBuffer::__ComplexityMasking()
{	
	unsigned char * base = (unsigned char *) __fa_bfr->GetBase();
    size_t length = __fa_bfr->size();
    _lowcomplexity_masker.Run( base, length, __bank.GetComplexity() );	
}
