/*
   $Id: BankBuffer.h 649 2011-10-31 13:57:42Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef _SBUFFER_H_
#define _SBUFFER_H_

#include "IndexTree.h"
#include "buffer.h"
#include "dust.h"
#include "LowComplexityMasker.h"

class Bank;
class Parameters;
class BlackList;

/**
 * \brief BankBuffer defines a buffer which will be filled with characters coming from the sequence to be indexed.
 *        The method IndexSequence() is called to read and index the sequence, filling the index_tree singleton
 */

class BankBuffer {
public:

    /* The constructors/destructor */
    BankBuffer(BlackList&,bool);
    ~BankBuffer();

    /* Reading and indexing the sequence whose filename is maintained
       by bank - Looping through the passes, calling __IndexSequenceFile() or __IndexSequenceMemory()
    */

    void IndexSequence();

    /* Filling the buffer: Putc puts a character, PutSeparator puts some $ characters
       Each one calls _Syncr() when the buffer is full
    */
private:
    Parameters& __parameters;
    Bank&       __bank;
	const Mask& __mask;
    Window      __window;
    IndexTree&  __index_tree;
    BlackList&  __black_list;
   
    void __Putc(char);
    void __PutSeparator() {__Putc('$');};

    /* Clearing the buffer */
    void __Clear();

	/* Resetting the position to 0  */
	void __ResetPos();
	
	/* Returning the starting position using the trick :  the starting pos is = -mas_length, hence 0 when the windo is full */
	uint32_t __GetStartingPos();

    /* querying the buffer
       IsFull tries to increase the buffer size until it reaches SBUFFER_MAX_SIZE
       Only when the buffer cannot be resized, true is returned
    */
    bool __IsFull() const {return __fa_bfr->IsFull();};
    bool __IsEmpty() const { return __fa_bfr->IsEmpty();};
    size_t __GetPosition() const {return __position;};

    //Buffer_ptr __GetBinaryIndexBankBuffer() const {return memory_mode ? (Buffer_ptr)bi_bfr : Buffer_ptr(NULL);}; // Strange here : (Buffer_ptr)bi_bfr was necessary for an icc compilation
    //Buffer_ptr __GetFaBankBuffer() const {return memory_mode ? (Buffer_ptr)fa_bfr : Buffer_ptr(NULL);};

    int32_t __position;       // Position of the top character, relative to the current indexed sequence (1st = 1)
    bool __memory_mode;    // if true, we don-t write the index to the files

    mutable Buffer_ptr __fa_bfr; // The buffer used to store the fa file
    mutable Buffer_ptr __bi_bfr; // The binary index, only used in memory_mode
    FILE* __fa_file;       // The sequence in internal 'fa' format - written during first pass

	/* Managing the complexity (i.e. Masking)
	 *   
	 * 
	 */
    LowComplexityMasker _lowcomplexity_masker;
 
    /* Reading and indexing a sequence
       Calling Putc or PutSeparator
       Calling _Syncr() just before returning
    */

    void __IndexSequenceFile();          // Called by IndexSequence when memory_mode==false
    void __IndexSequenceMemory();        // Called by IndexSequence when memory_mode==true
	void __ReadSeq(FILE *);

    Buffer* __InitFaBfr();     // Used by the constructor
    Buffer* __InitBiBfr();     // Used by the constructor

    void __Syncr();           // BankBuffer synchronization: dusting, indexing, dumping
    void __ComplexityMasking();  // Masking the regions of low complexity
    void __Index();           // Index the buffer, filling the tree and may be the blacklist
    void __Dump() {fwrite((char *)__fa_bfr->GetBase(),sizeof(char),__fa_bfr->size(),__fa_file);};  // Dump the buffer to fa_file
    void __ReadWord(uint32_t);   // Called by _Index or _IndexDinucleotides
    void __NewPosition(Leaf*,uint32_t); // Called by _Readword

};

/*
     The inline member functions of the class: BankBuffer
*/

/** 
 * \brief Clear the buffer
*/
inline void BankBuffer::__Clear()
{
    __fa_bfr->Clear();
}

inline void BankBuffer::__ResetPos()
{
    __position = 0;
}

inline uint32_t BankBuffer::__GetStartingPos()
{	
	// Equivalent to a rewind to set the position at the very first nucleotide (1) when the window if full (hence  __mask.GetLength())
    return (__position - __fa_bfr->size() - __mask.GetLength());  
}


/**
 * \brief put a character inside the buffer. 
 *        If the buffer gets full:
 *           - save the LAST WORD (mask size - 1) of the buffer
 *           - call __Syncr() (may be dusting, thus modifying the buffer)
 *           - Put the saved word at start of buffer
 * \param c The char to put
 */

inline void BankBuffer::__Putc(char c)
{
    if (c!='\n' && (signed char) c!= EOF)
    {
        // fa_bfr->Put must always be successfull, because we test the status with IsFull just after

        __fa_bfr->Put(c);
        __position++;
        if (__fa_bfr->IsFull())
			__Syncr();        
    }
}

#endif

