/*
   $Id: Display.cpp 649 2011-10-31 13:57:42Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/


#include <iostream>
using namespace std;

#include "utils.h"
#include "Display.h"
#include "QueryBuffer.h"
#include "rtstack.h"

/**
 * \brief A Hit is a set of hsps belonging to the same query sequence
 *          - Name is the name of the query sequence
 *          - score is the best hsp score
 *          - A vector of hsps is wrapped inside this structure
*/

struct Hit {
	    Hit(const string& n="", int16_t s=0) : score(s),Name(n) {};

		uint32_t score;
		string Name;
	    vector<Hsp*> hsps;
};

/**
 * \brief Use for sorting the hits: the key is the score, the hits are sorted in decreasing order of score
 */

inline bool operator< (const Hit& p, const Hit& q) { return p.score > q.score; }

/**
 * \brief A HitPile manages a list of hits
 * \todo ca n'est pas du tout une Pile !
 */

class HitPile {
    public:
	    HitPile() {};
	    void AddSortedHspPtr(Hsp* h);
	    void AddSortedUniqHspPtr(Hsp* h);
		uint32_t Size() const { return __hits.size(); };
	
	    /* Iterating through the hits */
	    typedef list<Hit>::const_iterator const_iterator;
	    const_iterator begin() const   {return __hits.begin();};
	    const_iterator end() const     {return __hits.end();};

	    /* Sorting the Hit pile */
	    void Sort() {__hits.sort();};

    private:
		list<Hit> __hits;
		void __NewHit ( Hsp* hsp_ptr)
		{
			Hit hit ( hsp_ptr->Subject.Name, hsp_ptr->GetScore());
			__hits.push_back( hit );
		}
};

/**
 * \brief Add the hsp_ptr to the last Hit, may be creating a new last Hit
 *        If the hsp_ptr as a Name different from the last Hit name, a new hit is created
 * \pre   The hsp must be sorted before continuously adding them
 * \param hsp_ptr a new hsp_ptr to add
 */

void HitPile::AddSortedHspPtr(Hsp* hsp_ptr)
{
	if ( __hits.empty() || hsp_ptr->Subject.Name != __hits.back().Name )
	{
		__NewHit( hsp_ptr );
	}
	__hits.back().hsps.push_back(hsp_ptr);
}

/**
 * \brief Add the hsp_ptr to the last Hit, may be creating a new last Hit
 *        The hsp is pushed ONLY if its score is == the hit score
 *        
 * \pre   The hsp must be sorted before continuously adding them
 * \param hsp_ptr a new hsp_ptr to add
 */

void HitPile::AddSortedUniqHspPtr(Hsp* hsp_ptr)
{
//	if ( __hits.empty() || hsp_ptr->Subject.Name != __hits.back().Name )
	if ( __hits.empty() )
	{
		__NewHit( hsp_ptr );
		__hits.back().hsps.push_back(hsp_ptr);
	}
	else
	{
		if ( hsp_ptr->GetScore() == __hits.back().score )
		{
			__hits.back().hsps.push_back(hsp_ptr);
		}
	}
}

/**
 * \brief Display a vector of Hsp, grouped by "hits"
 *        - Sort the hsps: increasing query name, then decreasing scores
 *        - Create a HitPile
 *        - Store the hsps inside the HitPile (see AddSortedHspPtr for the details)
 *        - Sort the HitPile (see operator> for the Hit)
 *        - Call _DisplayHitPile
 * 
 * \param v A vector of Hsp to be printed
 * \param len The number of hsp to print: all if -1, none if 0, else the specified number
 */

void Display::DisplayHits(vector<Hsp*>& v,int len) const
{
	if ( _query_ptr == NULL )
		throw ( logic_error ( (string) "ERROR - query unknown !" + ERROR_FILE_LINE(__FILE__,__LINE__)));

	HitPile hp;

    if (len==-1)
        len = v.size();
    if (len==0)
        return;

	// Sorting the hps, 1st key name, 2nd key score
    sort(v.begin(),v.end(),CompHspPtr);

    for (int i=0; i<len; i++)
    {
		if ( ! _parameters.IsBestScore() )
		{
			hp.AddSortedHspPtr(v[i]);
		}
		else
		{
			hp.AddSortedUniqHspPtr(v[i]);
		}
    };
    hp.Sort();
    _DisplayHitPile(hp);
}
	
/**
 * \brief Display a HitPile object: call _DisplayHeader, DisplayQueryInfo, then display the hits
 * 
 * \param hp The HitPile to display
 * 
 */

void Display::_DisplayHitPile(const HitPile& hp) const
{
    Hsp* first_hsp = *(hp.begin()->hsps.begin());

    _DisplayHeader(first_hsp->GetQueryName(),_parameters.GetBankName());
    DisplayQueryInfo(first_hsp->GetQueryName());
    for (HitPile::const_iterator i=hp.begin(); i!=hp.end(); ++i) {
    	__DisplaySingleHit(*i);
    }
}

/**
 * \brief Display a hit: Call DisplaySubjectInfo, then display the hsps
 * \param h The hit to display
 * 
 */

void Display::__DisplaySingleHit(const Hit& h) const
{
	DisplaySubjectInfo(h.Name,h.score);
    for (vector<Hsp*>::const_iterator j=h.hsps.begin(); j != h.hsps.end(); ++j) {
            DisplayHsp(**j);
    }
}

/**
 * \brief The DisplayM2 class implements the methods specific to the M2 output format  (see documentation)
*/

void DisplayM2::DisplayTopHeader() const
{
	fprintf(_dsp_file.Get(),"%s\n",M2_HEADER);
}

void DisplayM2::DisplayHsp(const Hsp& hsp) const
{
	char sign=(hsp.GetStrand() > 0) ? '+' : '-';
	fprintf(_dsp_file.Get(),"%s\t%u\t%u\t%s\t%u\t%u\t%c\t%d\t%d\n",hsp.GetQueryName().c_str(),hsp.GetQueryStart(),hsp.GetQueryEnd(),hsp.GetSubjectName().c_str(),hsp.GetSubjectStart(),hsp.GetSubjectEnd(),sign,hsp.GetScore(),hsp.GetSize());	
}

/**
 * \brief The DisplayM3 class implements the methods specific to the M3 output format  (see documentation)
*/

void DisplayM3::DisplayTopHeader() const
{
	fprintf(_dsp_file.Get(),"%s\n",M3_HEADER);
}

void DisplayM3::DisplayHsp(const Hsp& hsp) const
{

    const char *char_sign=(hsp.GetStrand() >0) ? OUTPUT_PLUS:OUTPUT_MINUS;

	fprintf(_dsp_file.Get(),"%s\t%s\t",hsp.GetQueryName().c_str(),hsp.GetSubjectName().c_str());
	fprintf(_dsp_file.Get(),"[%u-%u]->[%u-%u](%u,%s,%d,,%2.0f%c)",hsp.GetQueryStart(),hsp.GetQueryEnd(),hsp.GetSubjectStart(),hsp.GetSubjectEnd(),hsp.GetSize(),char_sign,hsp.GetScore(),hsp.GetPIdent(),'%');	
	fprintf(_dsp_file.Get(),"[");
    hsp.GetMismatch().Print(_dsp_file.Get());
	fprintf(_dsp_file.Get(),"]");
	fprintf(_dsp_file.Get(),"{");
    hsp.GetQueryGaps().Print(_dsp_file.Get());
	fprintf(_dsp_file.Get(),"}");
	fprintf(_dsp_file.Get(),"{");
    hsp.GetSubjectGaps().Print(_dsp_file.Get());
	fprintf(_dsp_file.Get(),"}:");
	fprintf(_dsp_file.Get(),"\n");
}

/**
 * \brief The DisplayM8 class implements the methods specific to the M8 output format  (see documentation)
*/

void DisplayM8::DisplayHsp(const Hsp& hsp) const
{
	uint32_t s_start,s_end;
	if ( hsp.GetStrand() > 0 )
	{
		s_start = hsp.GetSubjectStart();
		s_end   = hsp.GetSubjectEnd();
	}
	else
	{
		s_start = hsp.GetSubjectEnd();
		s_end   = hsp.GetSubjectStart();
	}
	fprintf(_dsp_file.Get(),"%s\t%s\t",hsp.GetQueryName().c_str(),hsp.GetSubjectName().c_str());
	fprintf(_dsp_file.Get(),"%2.2f\t%d\t%d\t%u\t%u\t%u\t%u\t%d\t%s\t%d\n",hsp.GetPIdent(),hsp.GetSize(),hsp.GetMismatch().GetSize(),hsp.GetGapOpenings().GetSize(),hsp.GetQueryStart(),hsp.GetQueryEnd(),s_start,s_end,"1e-0",hsp.GetScore());
}

/**
 * \brief The DisplayM9 class implements the methods specific to the M9 output format  (see documentation)
*/

void DisplayM9::_DisplayHeader(const string& query_name, const string& bank) const
{
	fprintf(_dsp_file.Get(),"# glint rel $Rev: 649 $ - Compiled on %s at %s\n", __DATE__, __TIME__);
	fprintf(_dsp_file.Get(),"# Query: %s\n",query_name.c_str());
	fprintf(_dsp_file.Get(),"# Bank: %s\n",bank.c_str());
	fprintf(_dsp_file.Get(),"# Fields: Query id, Subject id, %c identity, alignment length, mismatches, gap openings, q. start, q. end, s. start, s. end, e-value, bit score\n",'%');
}

/**
 * \brief The DisplayM4 class implements the methods specific to the M4 output format  (see documentation)
*/
void DisplayM4::DisplayHsp(const Hsp& hsp) const
{
	cout<<hsp<<endl;
}

/**
 * \brief The DisplayBed class implements the methods specific to the Bed output format  (see documentation)
*/

void DisplayBed::DisplayHsp(const Hsp& hsp) const
{
	string chrom1 = hsp.GetSubjectName();
	uint32_t  start1 = hsp.GetSubjectStart() - 1;  // Bed files are 0-based
	uint32_t  end1   = hsp.GetSubjectEnd();
	char   strand1= hsp.GetStrand()>0? '+' : '-';	

	// bed formatting
	if ( !_parameters.IsPairEnd())
	{
		uint32_t       score  = hsp.GetScore();
		string name        = hsp.GetQueryName();
		fprintf(_dsp_file.Get(),"%s\t%i\t%i\t%s\t%i\t%c\n",chrom1.c_str(),start1,end1,name.c_str(),score,strand1 );
	}	
	// bedpe formatting
	else if ( hsp.IsPaired() ) // if not we are not printing anything
	{
		string name     = GetReadName( hsp.GetQueryName() );  //The fragment extension (/1 and /2) is removed if any
		HspPtr mate_ptr = hsp.GetMate();
		string chrom2   = mate_ptr -> GetSubjectName();
		uint32_t start2    = mate_ptr -> GetSubjectStart() - 1; // Bed files are 0-based
		uint32_t end2      = mate_ptr -> GetSubjectEnd();
		//uint32_t score     = ((end1-start1)*hsp.GetScore() + (end2-start2)*mate_ptr -> GetScore()) / ( end1-start1 + end2-start2);
		uint32_t score     = hsp.GetScore() + mate_ptr -> GetScore(); //The score is simply the sum of the two scores
		char strand2      = mate_ptr -> GetStrand()>0 ? '+' : '-';
		fprintf(_dsp_file.Get(),"%s\t%i\t%i\t%s\t%i\t%i\t%s\t%i\t%c\t%c\n",chrom1.c_str(),start1,end1,chrom2.c_str(),start2,end2,name.c_str(),score,strand1,strand2 );	
	}
}

/**
 * \brief DisplayFasta - _DisplayHitPile calls DisplayHsp for the FIRST hsp only
 *
 * \param[in] hp The HitPile to print
 * 
 */

void DisplayFasta::_DisplayHitPile(const HitPile& hp) const
{	
	if ( _query_ptr == NULL )
		throw ( logic_error ( (string) "ERROR - query unknown !" + ERROR_FILE_LINE(__FILE__,__LINE__)));
		
	Hsp* h = *(hp.begin()->hsps.begin()); //The first hsp
	DisplayHsp(*h);
}

/**
 * \brief DisplayHsp - DisplayHsp only prints the fasta representation of the query sequence
 *
 * \param[in] h The Hsp to print
 * 
 */

void DisplayFasta::DisplayHsp(const Hsp & h) const
{	
	if ( _query_ptr == NULL )
		throw ( logic_error ( (string) "ERROR - query unknown !" + ERROR_FILE_LINE(__FILE__,__LINE__)));
		
	string SeqId = _query_ptr -> GetSeqId();
	fprintf(_dsp_file.Get(),">%s\tS=%d\tL=%d\tMM=%d\tG=%d\n",SeqId.c_str(),h.GetScore(),h.GetSize(),h.GetMismatch().GetSize(),h.GetQueryGaps().GetSize()+h.GetSubjectGaps().GetSize());
	fprintf(_dsp_file.Get(),"%s",_query_ptr -> GetSeqAsFasta() -> c_str());
}

/**
 * \brief DisplayFastq - _DisplayHitPile only calss DisplayHsp for the FIRST hsp
 *
 * \param[in] The HitPile to print
 * 
 */

void DisplayFastq::_DisplayHitPile(const HitPile& hp) const
{	
	if ( _query_ptr == NULL )
		throw ( logic_error ( (string) "ERROR - query unknown !" + ERROR_FILE_LINE(__FILE__,__LINE__)));
		
	Hsp* h = *(hp.begin()->hsps.begin()); //The first hsp
	DisplayHsp(*h);
}

/**
 * \brief DisplayFastq - DisplayHsp only prints the fastq representation of the query sequence
 *        NOTE - If the quality line is not conserved, we write a dummy quality just to be clean with the format
 *
 * \param[in] The HitPile to print
 * 
 */

void DisplayFastq::DisplayHsp(const Hsp& h) const
{	
	if ( _query_ptr == NULL )
		throw ( logic_error ( (string) "ERROR - query unknown !" + ERROR_FILE_LINE(__FILE__,__LINE__)));
		
	string SeqId =_query_ptr -> GetSeqId();
	fprintf(_dsp_file.Get(),"@%s\tS=%d\tL=%d\tMM=%d\tG=%d\n",SeqId.c_str(),h.GetScore(),h.GetSize(),h.GetMismatch().GetSize(),h.GetQueryGaps().GetSize()+h.GetSubjectGaps().GetSize());
	fprintf(_dsp_file.Get(),"%s",_query_ptr -> GetSeqOneLine()->c_str());
	fprintf(_dsp_file.Get(),"%s","\n+\n");
	if ( _query_ptr -> GetQual() == NULL )
	{
		for (unsigned int i=0; i< _query_ptr -> GetSeqLength(); ++i)
		{
			fputc('!',_dsp_file.Get());
		}
	}
	else
	{
		const char* q = _query_ptr -> GetQual();
		for (unsigned int i=0; i< _query_ptr -> GetSeqLength(); ++i)
		{
			fputc(*q++ + '!',_dsp_file.Get());
		}
	}		
	fputc('\n',_dsp_file.Get());
}

/**
 * \brief Display (fasta) - _DisplayHitPile only prints the fasta representation of the query sequence
 *
 * \param[in] The HitPile to print
 * 
 */

void DisplayFreq::DisplaySuitableFasta(const Buffer* freqBuffer, const stack<regionFreq>& suitableRegions) const
{
	string SeqId =_query_ptr -> GetSeqId();
	stack<regionFreq> regionTmp;
	regionTmp = suitableRegions;
	char* seq_base = (char*) _query_ptr -> GetBase();
	char* fbase    = (char*) freqBuffer -> GetBase();
	
	regionFreq r;
	int cptr = 1;
	while ( !regionTmp.empty() )
	{
		r = regionTmp.top();
		uint32_t j = 0;
		uint32_t i = 0;
		if( r.start < r.end )
		{
			fprintf( _dsp_file.Get(),">%d|%s\tstart : %d\tend : %d\n",cptr,SeqId.c_str(),r.start,r.end );
			for (j = r.start ; j < r.end ; ++j)
			{
				char a = *( fbase+j );
				char c = *( seq_base+j );
				if( a == 0 )
				{
					fprintf(_dsp_file.Get(),"%c",c);
					if( (i+1)%60 == 0 ) {fprintf(_dsp_file.Get(),"\n");}
				}
				else
				{
					fprintf(_dsp_file.Get(),"N");
					if( (i+1)%60 == 0 ) {fprintf(_dsp_file.Get(),"\n");}
				}
				++i;
			}
			if ( i%60 != 0)
			{
				fprintf(_dsp_file.Get(),"\n\n");
			}
		}
		regionTmp.pop();
		cptr++;
	}

	cout<<endl;
}
