/*
   $Id: scoring.cpp 648 2011-10-31 12:59:22Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

/*===========================================================================
                         scoring  -  description
                             -------------------
    begin				: septembre 2005
	 objet				: Functions for the initilizations of the scoring structures intended for dynamic programmming alignment-> alignment.c

=============================================================================*/


#include "scoring.h"

/**
 * \brief Initialize the scores for the dropoff and the dyn programming algorithm
 */ 

AlignConstants::AlignConstants() :
	__parameters(Parameters::Instance()),
	match    ( __parameters.GetMatchReward() ),
	mismatch ( __parameters.GetMisMatchPenality() ),
	nmatch   ( __parameters.GetNMatchPenality() ),
	__score  ( __InitScoringScheme (match,mismatch,nmatch)),
	__score_severe (__InitScoringSchemeSevere (match,mismatch,nmatch) ) {}

/**
 * \brief The destructor, releases memory
 */
AlignConstants::~AlignConstants()
{
	__Free( (void***) __score_severe);
	__Free( (void***) __score);
//	__FreeFromScoring( (void**) score_indel);
}


/**
 * \brief Free an array of arrays of arrays, 1st dimensions are 2 and 256
 */
void AlignConstants::__Free(void ***p)
{
	for (int s=0; s<2; ++s)
	{
		for (int i=0; i<256; ++i)
		{
			free ( p[s][i] );
		}
		free ( p[s] );
	}
	free ( p );
}

/**
 * \brief called by the constructor, allocate and initialize the scoring matrix for dynamic programming
 * \param rmatch, reward match
 * \param rmismatch reward, penality for a mismatch
 * \param nmatch, match against an N
 */
int ***AlignConstants::__InitScoringScheme(int rmatch, int rmismatch, int nmatch)
{
	int ***strand_matrix=(int***)MAlloc(2*sizeof(int**));
	for (int s01=0; s01<2; ++s01)
	{
		strand_matrix[s01] = __InitScoringSchemeStrand(rmatch, rmismatch, nmatch, s01);
	}
	return strand_matrix;
}
int ***AlignConstants::__InitScoringSchemeSevere(int rmatch, int rmismatch, int nmatch)
{
	int ***strand_matrix=(int***)MAlloc(2*sizeof(int**));
	for (int s01=0; s01<2; ++s01)
	{
		strand_matrix[s01] = __InitScoringSchemeSevereStrand(rmatch, rmismatch, nmatch, s01);
	}
	return strand_matrix;
}
/**
 * \brief called  by __InitScoringScheme and __InitScoringSchemeSevere , allocate the matrix for a given strand
 * \param rmatch, reward match
 * \param rmismatch reward, penality for a mismatch
 * \param nmatch, match against an N
 * \param strand01 (0 or 1)
 */
int **AlignConstants::__InitScoringSchemeStrand(int rmatch, int rmismatch, int nmatch, int strand01)
{
	int i,j;
	int **score_matrix=(int**)MAlloc(256*sizeof(int*));

	for (i=0;i<256;i++) {
		score_matrix[i]=(int*)MAlloc(256*sizeof(int));
		for (j=0;j<256;j++) {
			score_matrix[i][j]=rmismatch;
		}
	}
	for (i=0;i<256;i++) {
		score_matrix[i]['$']=N_INFTY;
		score_matrix['$'][i]=N_INFTY;
	}

	if (strand01 == 0) {
		score_matrix['A']['A']=rmatch;
		score_matrix['C']['C']=rmatch;
		score_matrix['G']['G']=rmatch;
		score_matrix['T']['T']=rmatch;
		score_matrix['a']['A']=rmatch;
		score_matrix['c']['C']=rmatch;
		score_matrix['g']['G']=rmatch;
		score_matrix['t']['T']=rmatch;
		score_matrix['A']['a']=rmatch;
		score_matrix['C']['c']=rmatch;
		score_matrix['G']['g']=rmatch;
		score_matrix['T']['t']=rmatch;
		score_matrix['a']['a']=rmatch;
		score_matrix['c']['c']=rmatch;
		score_matrix['g']['g']=rmatch;
		score_matrix['t']['t']=rmatch;
		
		//Incorporating IUPAC codes
		score_matrix['R']['A']=rmatch;
		score_matrix['R']['G']=rmatch;
		score_matrix['Y']['C']=rmatch;
		score_matrix['Y']['T']=rmatch;
		score_matrix['W']['A']=rmatch;
		score_matrix['W']['T']=rmatch;
		score_matrix['S']['C']=rmatch;
		score_matrix['S']['T']=rmatch;
		score_matrix['M']['A']=rmatch;
		score_matrix['M']['C']=rmatch;
		score_matrix['K']['G']=rmatch;
		score_matrix['K']['T']=rmatch;
		score_matrix['B']['C']=rmatch;
		score_matrix['B']['G']=rmatch;
		score_matrix['B']['T']=rmatch;
		score_matrix['D']['A']=rmatch;
		score_matrix['D']['G']=rmatch;
		score_matrix['D']['T']=rmatch;
		score_matrix['H']['A']=rmatch;
		score_matrix['H']['C']=rmatch;
		score_matrix['H']['T']=rmatch;
		score_matrix['H']['A']=rmatch;
		score_matrix['H']['C']=rmatch;
		score_matrix['H']['G']=rmatch;
	} else {
		score_matrix['A']['T']=rmatch;
		score_matrix['T']['A']=rmatch;
		score_matrix['C']['G']=rmatch;
		score_matrix['G']['C']=rmatch;
		score_matrix['A']['t']=rmatch;
		score_matrix['T']['a']=rmatch;
		score_matrix['C']['g']=rmatch;
		score_matrix['G']['c']=rmatch;
		score_matrix['a']['T']=rmatch;
		score_matrix['t']['A']=rmatch;
		score_matrix['c']['G']=rmatch;
		score_matrix['g']['C']=rmatch;
		score_matrix['a']['t']=rmatch;
		score_matrix['t']['a']=rmatch;
		score_matrix['c']['g']=rmatch;
		score_matrix['g']['c']=rmatch;
		
		//Incorporating IUPAC codes
		score_matrix['R']['T']=rmatch;
		score_matrix['R']['C']=rmatch;
		score_matrix['Y']['G']=rmatch;
		score_matrix['Y']['A']=rmatch;
		score_matrix['W']['T']=rmatch;
		score_matrix['W']['A']=rmatch;
		score_matrix['S']['G']=rmatch;
		score_matrix['S']['A']=rmatch;
		score_matrix['M']['T']=rmatch;
		score_matrix['M']['G']=rmatch;
		score_matrix['K']['C']=rmatch;
		score_matrix['K']['A']=rmatch;
		score_matrix['B']['G']=rmatch;
		score_matrix['B']['C']=rmatch;
		score_matrix['B']['A']=rmatch;
		score_matrix['D']['T']=rmatch;
		score_matrix['D']['C']=rmatch;
		score_matrix['D']['A']=rmatch;
		score_matrix['H']['T']=rmatch;
		score_matrix['H']['G']=rmatch;
		score_matrix['H']['A']=rmatch;
		score_matrix['H']['T']=rmatch;
		score_matrix['H']['G']=rmatch;
		score_matrix['H']['C']=rmatch;		
	}

	/*score_matrix['N']['A']=0;
	*/
	score_matrix['A']['N']=score_matrix['N']['A']=nmatch;
	score_matrix['C']['N']=score_matrix['N']['C']=nmatch;
	score_matrix['G']['N']=score_matrix['N']['G']=nmatch;
	score_matrix['T']['N']=score_matrix['N']['T']=nmatch;
	score_matrix['N']['N']=nmatch;

	return score_matrix;
}

/**
 * \brief allocate and initialize a severe scoring matrix for dynamic programming
 *        (used when the Strict option is set to true)
 *        lower case letters matches are treated as N letter matches (no reward)
 * \param rmatch, reward match
 * \param rmismatch reward, penality for a mismatch
 * \param nmatch, match against an N
 * \param strand01 (+1 or -1)
 */
int **AlignConstants::__InitScoringSchemeSevereStrand(int rmatch, int rmismatch, int nmatch, int strand01)
{
	int i,j;
	int **score_matrix=(int**)MAlloc(256*sizeof(int*));

	for (i=0;i<256;i++) {
		score_matrix[i]=(int*)MAlloc(256*sizeof(int));
		for (j=0;j<256;j++) {
			score_matrix[i][j]=rmismatch;
		}
	}
	for (i=0;i<256;i++) {
		score_matrix[i]['$']=N_INFTY;
		score_matrix['$'][i]=N_INFTY;
	}

	if (strand01 == 0) {
		score_matrix['A']['A']=rmatch;
		score_matrix['C']['C']=rmatch;
		score_matrix['G']['G']=rmatch;
		score_matrix['T']['T']=rmatch;
		score_matrix['a']['A']=nmatch;
		score_matrix['c']['C']=nmatch;
		score_matrix['g']['G']=nmatch;
		score_matrix['t']['T']=nmatch;
		score_matrix['A']['a']=nmatch;
		score_matrix['C']['c']=nmatch;
		score_matrix['G']['g']=nmatch;
		score_matrix['T']['t']=nmatch;
		score_matrix['a']['a']=nmatch;
		score_matrix['c']['c']=nmatch;
		score_matrix['g']['g']=nmatch;
		score_matrix['t']['t']=nmatch;
	} else {
		score_matrix['A']['T']=rmatch;
		score_matrix['T']['A']=rmatch;
		score_matrix['C']['G']=rmatch;
		score_matrix['G']['C']=rmatch;
		score_matrix['A']['t']=nmatch;
		score_matrix['T']['a']=nmatch;
		score_matrix['C']['g']=nmatch;
		score_matrix['G']['c']=nmatch;
		score_matrix['a']['T']=nmatch;
		score_matrix['t']['A']=nmatch;
		score_matrix['c']['G']=nmatch;
		score_matrix['g']['C']=nmatch;
		score_matrix['a']['t']=nmatch;
		score_matrix['t']['a']=nmatch;
		score_matrix['c']['g']=nmatch;
		score_matrix['g']['c']=nmatch;
	}

	/*score_matrix['N']['A']=0;
	*/
	score_matrix['A']['N']=score_matrix['N']['A']=nmatch;
	score_matrix['C']['N']=score_matrix['N']['C']=nmatch;
	score_matrix['G']['N']=score_matrix['N']['G']=nmatch;
	score_matrix['T']['N']=score_matrix['N']['T']=nmatch;
	score_matrix['N']['N']=nmatch;

	return score_matrix;
}

/**
 * \brief called by the constructor, allocate and initialize a scoring matrix indels
 * \param alpha The gap penalty (???)
 * \param beta NOT USED ???
 */
#ifdef REMOVED
int **AlignConstants::__InitScoringIndels(int alpha, int beta)
{
	int i,j;
	int **score_matrix=(int**)MAlloc(256*sizeof(int*));

	for (i=0;i<256;i++) {
		score_matrix[i]=(int*)MAlloc(256*sizeof(int));
		for (j=0;j<256;j++) {
			score_matrix[i][j]=alpha;
		}
	}
	for (i=0;i<256;i++) {
		score_matrix[i]['$']=N_INFTY;
		score_matrix['$'][i]=N_INFTY;
	}

	return score_matrix;
}
#endif
/*----------------------------------------------*/

int __Overlap(const Mask& m, int i, int j, int k)
{
	int r;
	int score=0;

	for (r=k;r<m.GetLength();r++) {
		if (m[r-k]<=j && m[r]>i) score++;
	}
	for (r=m.GetLength();r<m.GetLength()+k;r++) {
		if (m[r-k]<=j) score++;
	}

	return score;
}


int __Overlap_bw(const Mask& m, int i, int j, int k)
{
	int r;
	int score=0;

	for (r=0;r<k;r++) {
		if (m[r]<=i) score++;
	}

	for (r=k;r<m.GetLength();r++) {
		if (m[r]<=i && m[r-k]>j) score++;
	}

	return score;
}

/**
 * \brief called by the constructor, allocate and initialize a scoring matrix for the overlaps
 */

short *** MatchConstants::__InitOverlapScore(int (overlap)(const Mask&, int, int, int k))
{
	size_t i,j,k;
	size_t leaf_level  = (size_t) __mask.GetLeafLevel();
	size_t mask_length = (size_t) __mask.GetLength();
	
	short ***overlap_score=(short***)MAlloc((leaf_level+1)*sizeof(short**));

	//cout<<"mask : "<<m.GetString()<<" taille "<<m.GetLength()<<"\n";

	for (i=ROOT_LEVEL;i<=leaf_level;i++) {
		overlap_score[i]=(short**)MAlloc((leaf_level+1)*sizeof(short*));
		for (j=ROOT_LEVEL;j<=leaf_level;j++) {
			overlap_score[i][j]=(short*)MAlloc(mask_length*sizeof(short));
			for (k=0;k<mask_length;k++) {
				overlap_score[i][j][k]=overlap(__mask,i,j,k);
				//fprintf(stdout,"hit(i)=%d hit(i+1)=%d decalage=%d score=%d\n",i,j,k,overlap_score[i][j][k]);
			}
		}
	}
	return overlap_score;
}

void MatchConstants::__FreeFromScoring ( short*** p )
{
	size_t leaf_level  = (size_t) __mask.GetLeafLevel();
	for ( size_t i=ROOT_LEVEL;i<=leaf_level;i++ )
	{
		for ( size_t j=ROOT_LEVEL;j<=leaf_level;j++ )
		{
			free (p[i][j]);
		}
		free (p[i]);
	}
	free ( p );
}
